@@ -1,6 +1,6 @@
 
 /*
- * $Id: neighbors.c,v 1.299.2.4 2003/11/29 18:53:00 hno Exp $
+ * $Id: neighbors.c,v 1.312 2006/12/10 05:18:47 hno Exp $
  *
  * DEBUG: section 15    Neighbor Routines
  * AUTHOR: Harvest Derived
@@ -38,7 +38,6 @@
 /* count mcast group peers every 15 minutes */
 #define MCAST_COUNT_RATE 900
 
-static int peerAllowedToUse(const peer *, request_t *);
 static int peerWouldBePinged(const peer *, request_t *);
 static void neighborRemove(peer *);
 static void neighborAlive(peer *, const MemObject *, const icp_common_t *);
@@ -114,7 +113,7 @@ neighborType(const peer * p, const request_t * request)
  * this function figures out if it is appropriate to fetch REQUEST
  * from PEER.
  */
-static int
+int
 peerAllowedToUse(const peer * p, request_t * request)
 {
     const struct _domain_ping *d = NULL;
@@ -266,8 +265,12 @@ getRoundRobinParent(request_t * request)
 	    continue;
 	if (!peerHTTPOkay(p, request))
 	    continue;
-	if (q && q->rr_count < p->rr_count)
+	if (p->weight == 1) {
+	    if (q && q->rr_count < p->rr_count)
+		continue;
+	} else if (p->weight == 0 || (q && q->rr_count < (p->rr_count / p->weight))) {
 	    continue;
+	}
 	q = p;
     }
     if (q)
@@ -362,7 +365,7 @@ neighborRemove(peer * target)
 }
 
 void
-neighbors_open(int fd)
+neighbors_init(void)
 {
     struct sockaddr_in name;
     socklen_t len = sizeof(struct sockaddr_in);
@@ -370,25 +373,27 @@ neighbors_open(int fd)
     const char *me = getMyHostname();
     peer *this;
     peer *next;
-    memset(&name, '\0', sizeof(struct sockaddr_in));
-    if (getsockname(fd, (struct sockaddr *) &name, &len) < 0)
-	debug(15, 1) ("getsockname(%d,%p,%p) failed.\n", fd, &name, &len);
-    for (this = Config.peers; this; this = next) {
-	sockaddr_in_list *s;
-	next = this->next;
-	if (0 != strcmp(this->host, me))
-	    continue;
-	for (s = Config.Sockaddr.http; s; s = s->next) {
-	    if (this->http_port != ntohs(s->s.sin_port))
+    int fd = theInIcpConnection;
+    if (fd >= 0) {
+	memset(&name, '\0', sizeof(struct sockaddr_in));
+	if (getsockname(fd, (struct sockaddr *) &name, &len) < 0)
+	    debug(15, 1) ("getsockname(%d,%p,%p) failed.\n", fd, &name, &len);
+	for (this = Config.peers; this; this = next) {
+	    http_port_list *s;
+	    next = this->next;
+	    if (0 != strcmp(this->host, me))
 		continue;
-	    debug(15, 1) ("WARNING: Peer looks like this host\n");
-	    debug(15, 1) ("         Ignoring %s %s/%d/%d\n",
-		neighborTypeStr(this), this->host, this->http_port,
-		this->icp.port);
-	    neighborRemove(this);
+	    for (s = Config.Sockaddr.http; s; s = s->next) {
+		if (this->http_port != ntohs(s->s.sin_port))
+		    continue;
+		debug(15, 1) ("WARNING: Peer looks like this host\n");
+		debug(15, 1) ("         Ignoring %s %s/%d/%d\n",
+		    neighborTypeStr(this), this->host, this->http_port,
+		    this->icp.port);
+		neighborRemove(this);
+	    }
 	}
     }
-
     peerRefreshDNS((void *) 1);
     if (0 == echo_hdr.opcode) {
 	echo_hdr.opcode = ICP_SECHO;
@@ -405,9 +410,11 @@ neighbors_open(int fd)
     cachemgrRegister("server_list",
 	"Peer Cache Statistics",
 	neighborDumpPeers, 0, 1);
-    cachemgrRegister("non_peers",
-	"List of Unknown sites sending ICP messages",
-	neighborDumpNonPeers, 0, 1);
+    if (theInIcpConnection >= 0) {
+	cachemgrRegister("non_peers",
+	    "List of Unknown sites sending ICP messages",
+	    neighborDumpNonPeers, 0, 1);
+    }
 }
 
 int
@@ -485,11 +492,6 @@ neighborsUdpPing(request_t * request,
 
 	p->stats.pings_sent++;
 	if (p->type == PEER_MULTICAST) {
-	    /*
-	     * set a bogus last_reply time so neighborUp() never
-	     * says a multicast peer is dead.
-	     */
-	    p->stats.last_reply = squid_curtime;
 	    mcast_exprep += p->mcast.n_replies_expected;
 	    mcast_timeout += (p->stats.rtt * p->mcast.n_replies_expected);
 	} else if (neighborUp(p)) {
@@ -505,14 +507,17 @@ neighborsUdpPing(request_t * request,
 	    /* Neighbor is dead; ping it anyway, but don't expect a reply */
 	    /* log it once at the threshold */
 	    if (p->stats.logged_state == PEER_ALIVE) {
-		debug(15, 1) ("Detected DEAD %s: %s/%d/%d\n",
-		    neighborTypeStr(p),
-		    p->host, p->http_port, p->icp.port);
+		debug(15, 1) ("Detected DEAD %s: %s\n",
+		    neighborTypeStr(p), p->name);
 		p->stats.logged_state = PEER_DEAD;
 	    }
 	}
 	p->stats.last_query = squid_curtime;
-	if (p->stats.probe_start == 0)
+	/*
+	 * keep probe_start == 0 for a multicast peer,
+	 * so neighborUp() never says this peer is dead.
+	 */
+	if ((p->type != PEER_MULTICAST) && (p->stats.probe_start == 0))
 	    p->stats.probe_start = squid_curtime;
     }
     if ((first_ping = first_ping->next) == NULL)
@@ -589,6 +594,9 @@ peerDigestLookup(peer * p, request_t * request)
     if (!p->digest) {
 	debug(15, 5) ("peerDigestLookup: gone!\n");
 	return LOOKUP_NONE;
+    } else if (!peerHTTPOkay(p, request)) {
+	debug(15, 5) ("peerDigestLookup: !peerHTTPOkay\n");
+	return LOOKUP_NONE;
     } else if (!p->digest->flags.needed) {
 	debug(15, 5) ("peerDigestLookup: note need\n");
 	peerDigestNeeded(p->digest);
@@ -597,9 +605,6 @@ peerDigestLookup(peer * p, request_t * request)
 	debug(15, 5) ("peerDigestLookup: !ready && %srequested\n",
 	    p->digest->flags.requested ? "" : "!");
 	return LOOKUP_NONE;
-    } else if (!peerHTTPOkay(p, request)) {
-	debug(15, 5) ("peerDigestLookup: !peerHTTPOkay\n");
-	return LOOKUP_NONE;
     }
     debug(15, 5) ("peerDigestLookup: OK to lookup peer %s\n", p->host);
     assert(p->digest->cd);
@@ -681,9 +686,8 @@ static void
 neighborAlive(peer * p, const MemObject * mem, const icp_common_t * header)
 {
     if (p->stats.logged_state == PEER_DEAD && p->tcp_up) {
-	debug(15, 1) ("Detected REVIVED %s: %s/%d/%d\n",
-	    neighborTypeStr(p),
-	    p->host, p->http_port, p->icp.port);
+	debug(15, 1) ("Detected REVIVED %s: %s\n",
+	    neighborTypeStr(p), p->name);
 	p->stats.logged_state = PEER_ALIVE;
     }
     p->stats.last_reply = squid_curtime;
@@ -714,9 +718,8 @@ static void
 neighborAliveHtcp(peer * p, const MemObject * mem, const htcpReplyData * htcp)
 {
     if (p->stats.logged_state == PEER_DEAD && p->tcp_up) {
-	debug(15, 1) ("Detected REVIVED %s: %s/%d/%d\n",
-	    neighborTypeStr(p),
-	    p->host, p->http_port, p->icp.port);
+	debug(15, 1) ("Detected REVIVED %s: %s\n",
+	    neighborTypeStr(p), p->name);
 	p->stats.logged_state = PEER_ALIVE;
     }
     p->stats.last_reply = squid_curtime;
@@ -907,7 +910,7 @@ peerFindByName(const char *name)
 {
     peer *p = NULL;
     for (p = Config.peers; p; p = p->next) {
-	if (!strcasecmp(name, p->host))
+	if (!strcasecmp(name, p->name))
 	    break;
     }
     return p;
@@ -918,7 +921,7 @@ peerFindByNameAndPort(const char *name, unsigned short port)
 {
     peer *p = NULL;
     for (p = Config.peers; p; p = p->next) {
-	if (strcasecmp(name, p->host))
+	if (strcasecmp(name, p->name))
 	    continue;
 	if (port != p->http_port)
 	    continue;
@@ -934,10 +937,9 @@ neighborUp(const peer * p)
 	if (!peerProbeConnect((peer *) p))
 	    return 0;
     }
-    if (p->options.no_query)
-	return 1;
-    if (p->stats.probe_start != 0 &&
-	squid_curtime - p->stats.probe_start > Config.Timeout.deadPeer)
+    if (p->stats.logged_state != PEER_ALIVE)
+	return 0;
+    if (p->monitor.state != PEER_ALIVE)
 	return 0;
     /*
      * The peer can not be UP if we don't have any IP addresses
@@ -945,6 +947,11 @@ neighborUp(const peer * p)
      */
     if (0 == p->n_addresses)
 	return 0;
+    if (p->options.no_query)
+	return 1;
+    if (p->stats.probe_start != 0 &&
+	squid_curtime - p->stats.probe_start > Config.Timeout.deadPeer)
+	return 0;
     return 1;
 }
 
@@ -961,7 +968,10 @@ peerDestroy(void *data)
 	safe_free(l->domain);
 	safe_free(l);
     }
+    aclDestroyAccessList(&p->access);
     safe_free(p->host);
+    safe_free(p->name);
+    safe_free(p->domain);
 #if USE_CACHE_DIGESTS
     if (p->digest) {
 	PeerDigest *pd = p->digest;
@@ -1032,8 +1042,9 @@ peerRefreshDNS(void *data)
 	eventAddIsh("peerRefreshDNS", peerRefreshDNS, NULL, 180.0, 1);
 	return;
     }
-    for (p = Config.peers; p; p = p->next)
+    for (p = Config.peers; p; p = p->next) {
 	ipcache_nbgethostbyname(p->host, peerDNSConfigure, p);
+    }
     /* Reconfigure the peers every hour */
     eventAddIsh("peerRefreshDNS", peerRefreshDNS, NULL, 3600.0, 1);
 }
@@ -1048,9 +1059,8 @@ peerConnectFailedSilent(peer * p)
     }
     p->tcp_up--;
     if (!p->tcp_up) {
-	debug(15, 1) ("Detected DEAD %s: %s/%d/%d\n",
-	    neighborTypeStr(p),
-	    p->host, p->http_port, p->icp.port);
+	debug(15, 1) ("Detected DEAD %s: %s\n",
+	    neighborTypeStr(p), p->name);
 	p->stats.logged_state = PEER_DEAD;
     }
 }
@@ -1067,9 +1077,9 @@ peerConnectSucceded(peer * p)
 {
     if (!p->tcp_up) {
 	debug(15, 2) ("TCP connection to %s/%d succeded\n", p->host, p->http_port);
-	debug(15, 1) ("Detected REVIVED %s: %s/%d/%d\n",
-	    neighborTypeStr(p),
-	    p->host, p->http_port, p->icp.port);
+	debug(15, 1) ("Detected REVIVED %s: %s\n",
+	    neighborTypeStr(p), p->name);
+	peerMonitorNow(p);
 	p->stats.logged_state = PEER_ALIVE;
     }
     p->tcp_up = PEER_TCP_MAGIC_COUNT;
@@ -1098,7 +1108,7 @@ peerProbeConnect(peer * p)
 	return ret;		/* probe already running */
     if (squid_curtime - p->stats.last_connect_probe == 0)
 	return ret;		/* don't probe to often */
-    fd = comm_open(SOCK_STREAM, 0, getOutgoingAddr(NULL),
+    fd = comm_open(SOCK_STREAM, IPPROTO_TCP, getOutgoingAddr(NULL),
 	0, COMM_NONBLOCKING, p->host);
     if (fd < 0)
 	return ret;
@@ -1158,6 +1168,7 @@ peerCountMcastPeersStart(void *data)
     psstate->entry = fake;
     psstate->callback = NULL;
     psstate->callback_data = p;
+    cbdataLock(psstate->callback_data);
     psstate->ping.start = current_time;
     mem = fake->mem_obj;
     mem->request = requestLink(psstate->request);
@@ -1177,7 +1188,7 @@ peerCountMcastPeersStart(void *data)
     eventAdd("peerCountMcastPeersDone",
 	peerCountMcastPeersDone,
 	psstate,
-	(double) Config.Timeout.mcast_icp_query, 1);
+	Config.Timeout.mcast_icp_query / 1000.0, 1);
     p->mcast.flags.counting = 1;
     peerCountMcastPeersSchedule(p, MCAST_COUNT_RATE);
 }
@@ -1188,17 +1199,20 @@ peerCountMcastPeersDone(void *data)
     ps_state *psstate = data;
     peer *p = psstate->callback_data;
     StoreEntry *fake = psstate->entry;
-    p->mcast.flags.counting = 0;
-    p->mcast.avg_n_members = doubleAverage(p->mcast.avg_n_members,
-	(double) psstate->ping.n_recv,
-	++p->mcast.n_times_counted,
-	10);
-    debug(15, 1) ("Group %s: %d replies, %4.1f average, RTT %d\n",
-	p->host,
-	psstate->ping.n_recv,
-	p->mcast.avg_n_members,
-	p->stats.rtt);
-    p->mcast.n_replies_expected = (int) p->mcast.avg_n_members;
+    if (cbdataValid(p)) {
+	p->mcast.flags.counting = 0;
+	p->mcast.avg_n_members = doubleAverage(p->mcast.avg_n_members,
+	    (double) psstate->ping.n_recv,
+	    ++p->mcast.n_times_counted,
+	    10);
+	debug(15, 1) ("Group %s: %d replies, %4.1f average, RTT %d\n",
+	    p->host,
+	    psstate->ping.n_recv,
+	    p->mcast.avg_n_members,
+	    p->stats.rtt);
+	p->mcast.n_replies_expected = (int) p->mcast.avg_n_members;
+    }
+    cbdataUnlock(p);
     EBIT_SET(fake->flags, ENTRY_ABORTED);
     requestUnlink(fake->mem_obj->request);
     fake->mem_obj->request = NULL;
@@ -1249,8 +1263,14 @@ dump_peer_options(StoreEntry * sentry, peer * p)
 	storeAppendPrintf(sentry, " round-robin");
     if (p->options.mcast_responder)
 	storeAppendPrintf(sentry, " multicast-responder");
+    if (p->weight != 1)
+	storeAppendPrintf(sentry, " weight=%d", p->weight);
     if (p->options.closest_only)
 	storeAppendPrintf(sentry, " closest-only");
+    if (p->options.userhash)
+	storeAppendPrintf(sentry, " userhash");
+    if (p->options.sourcehash)
+	storeAppendPrintf(sentry, " sourcehash");
 #if USE_HTCP
     if (p->options.htcp)
 	storeAppendPrintf(sentry, " htcp");
@@ -1265,6 +1285,29 @@ dump_peer_options(StoreEntry * sentry, peer * p)
 	storeAppendPrintf(sentry, " login=%s", p->login);
     if (p->mcast.ttl > 0)
 	storeAppendPrintf(sentry, " ttl=%d", p->mcast.ttl);
+    if (p->connect_timeout > 0)
+	storeAppendPrintf(sentry, " connect-timeout=%d", (int) p->connect_timeout);
+#if USE_CACHE_DIGESTS
+    if (p->digest_url)
+	storeAppendPrintf(sentry, " digest-url=%s", p->digest_url);
+#endif
+    if (p->options.allow_miss)
+	storeAppendPrintf(sentry, " allow-miss");
+    if (p->max_conn > 0)
+	storeAppendPrintf(sentry, " max-conn=%d", p->max_conn);
+    if (p->options.originserver)
+	storeAppendPrintf(sentry, " originserver");
+    /* name is used in the heading */
+    if (p->monitor.url)
+	storeAppendPrintf(sentry, " monitorurl=%s", p->monitor.url);
+    if (p->monitor.min > 0 || p->monitor.max > 0) {
+	if (p->monitor.max > 0)
+	    storeAppendPrintf(sentry, " monitorsize=%d:%d", p->monitor.min, p->monitor.max);
+	else
+	    storeAppendPrintf(sentry, " monitorsize=%d", p->monitor.min);
+    }
+    if (p->domain)
+	storeAppendPrintf(sentry, " forceddomain=%s", p->domain);
     storeAppendPrintf(sentry, "\n");
 }
 
@@ -1279,8 +1322,10 @@ dump_peers(StoreEntry * sentry, peer * peers)
 	storeAppendPrintf(sentry, "There are no neighbors installed.\n");
     for (e = peers; e; e = e->next) {
 	assert(e->host != NULL);
-	storeAppendPrintf(sentry, "\n%-11.11s: %s/%d/%d\n",
+	storeAppendPrintf(sentry, "\n%-11.11s: %s\n",
 	    neighborTypeStr(e),
+	    e->name);
+	storeAppendPrintf(sentry, "Host       : %s/%d/%d\n",
 	    e->host,
 	    e->http_port,
 	    e->icp.port);
@@ -1294,42 +1339,49 @@ dump_peers(StoreEntry * sentry, peer * peers)
 	    neighborUp(e) ? "Up" : "Down");
 	storeAppendPrintf(sentry, "AVG RTT    : %d msec\n", e->stats.rtt);
 	storeAppendPrintf(sentry, "OPEN CONNS : %d\n", e->stats.conn_open);
-	storeAppendPrintf(sentry, "LAST QUERY : %8d seconds ago\n",
-	    (int) (squid_curtime - e->stats.last_query));
-	storeAppendPrintf(sentry, "LAST REPLY : %8d seconds ago\n",
-	    (int) (squid_curtime - e->stats.last_reply));
-	storeAppendPrintf(sentry, "PINGS SENT : %8d\n", e->stats.pings_sent);
-	storeAppendPrintf(sentry, "PINGS ACKED: %8d %3d%%\n",
-	    e->stats.pings_acked,
-	    percent(e->stats.pings_acked, e->stats.pings_sent));
+	if (!e->options.no_query) {
+	    storeAppendPrintf(sentry, "LAST QUERY : %8d seconds ago\n",
+		(int) (squid_curtime - e->stats.last_query));
+	    if (e->stats.last_reply > 0)
+		storeAppendPrintf(sentry, "LAST REPLY : %8d seconds ago\n",
+		    (int) (squid_curtime - e->stats.last_reply));
+	    else
+		storeAppendPrintf(sentry, "LAST REPLY : none received\n");
+	    storeAppendPrintf(sentry, "PINGS SENT : %8d\n", e->stats.pings_sent);
+	    storeAppendPrintf(sentry, "PINGS ACKED: %8d %3d%%\n",
+		e->stats.pings_acked,
+		percent(e->stats.pings_acked, e->stats.pings_sent));
+	}
 	storeAppendPrintf(sentry, "FETCHES    : %8d %3d%%\n",
 	    e->stats.fetches,
 	    percent(e->stats.fetches, e->stats.pings_acked));
 	storeAppendPrintf(sentry, "IGNORED    : %8d %3d%%\n",
 	    e->stats.ignored_replies,
 	    percent(e->stats.ignored_replies, e->stats.pings_acked));
-	storeAppendPrintf(sentry, "Histogram of PINGS ACKED:\n");
+	if (!e->options.no_query) {
+	    storeAppendPrintf(sentry, "Histogram of PINGS ACKED:\n");
 #if USE_HTCP
-	if (e->options.htcp) {
-	    storeAppendPrintf(sentry, "\tMisses\t%8d %3d%%\n",
-		e->htcp.counts[0],
-		percent(e->htcp.counts[0], e->stats.pings_acked));
-	    storeAppendPrintf(sentry, "\tHits\t%8d %3d%%\n",
-		e->htcp.counts[1],
-		percent(e->htcp.counts[1], e->stats.pings_acked));
-	} else {
+	    if (e->options.htcp) {
+		storeAppendPrintf(sentry, "\tMisses\t%8d %3d%%\n",
+		    e->htcp.counts[0],
+		    percent(e->htcp.counts[0], e->stats.pings_acked));
+		storeAppendPrintf(sentry, "\tHits\t%8d %3d%%\n",
+		    e->htcp.counts[1],
+		    percent(e->htcp.counts[1], e->stats.pings_acked));
+	    } else {
 #endif
-	    for (op = ICP_INVALID; op < ICP_END; op++) {
-		if (e->icp.counts[op] == 0)
-		    continue;
-		storeAppendPrintf(sentry, "    %12.12s : %8d %3d%%\n",
-		    icp_opcode_str[op],
-		    e->icp.counts[op],
-		    percent(e->icp.counts[op], e->stats.pings_acked));
+		for (op = ICP_INVALID; op < ICP_END; op++) {
+		    if (e->icp.counts[op] == 0)
+			continue;
+		    storeAppendPrintf(sentry, "    %12.12s : %8d %3d%%\n",
+			icp_opcode_str[op],
+			e->icp.counts[op],
+			percent(e->icp.counts[op], e->stats.pings_acked));
+		}
+#if USE_HTCP
 	    }
-#if USE_HTCP
-	}
 #endif
+	}
 	if (e->stats.last_connect_failure) {
 	    storeAppendPrintf(sentry, "Last failed connect() at: %s\n",
 		mkhttpdlogtime(&(e->stats.last_connect_failure)));
